/****************************************************************************/
/*              Beebem - (c) David Alan Gilbert 1994                        */
/*              ------------------------------------                        */
/* This program may be distributed freely within the following restrictions:*/
/*                                                                          */
/* 1) You may not charge for this program or for any part of it.            */
/* 2) This copyright message must be distributed with all copies.           */
/* 3) This program must be distributed complete with source code.  Binary   */
/*    only distribution is not permitted.                                   */
/* 4) The author offers no warrenties, or guarentees etc. - you use it at   */
/*    your own risk.  If it messes something up or destroys your computer   */
/*    thats YOUR problem.                                                   */
/* 5) You may use small sections of code from this program in your own      */
/*    applications - but you must acknowledge its use.  If you plan to use  */
/*    large sections then please ask the author.                            */
/*                                                                          */
/* If you do not agree with any of the above then please do not use this    */
/* program.                                                                 */
/* Please report any problems to the author at beebem@treblig.org           */
/****************************************************************************/
/* Video handling -          David Alan Gilbert */

/* Version 2 - 24/12/94 - Designed to emulate start of frame interrupt
   correctly */

/* Mike Wyatt 7/6/97 - Added cursor display and Win32 port */

#include <iostream>
#include <stdlib.h>
#include <SDL.h>

#include "6502core.h"
#include "beebmem.h"
#include "beebwin.h"
#include "main.h"
#include "mode7font.h"
#include "sysvia.h"
#include "video.h"

#include "beebconfig.h"
#include "beebconfig_data.h"

#include "uefstate.h"

#include "virtualkeyboard.h"

#ifdef BEEB_DOTIME
#include <sys/times.h>
#ifdef SUNOS
#include <sys/param.h>
#endif
#ifdef HPUX
#include <unistd.h>
#endif
#endif


/* Handle SDL surface locking:
 */
extern SDL_Surface *frame_buffer_p;
#define LOCK(s)   {if(SDL_MUSTLOCK(s))(void)SDL_LockSurface(s);}
#define UNLOCK(s) {if(SDL_MUSTLOCK(s))(void)SDL_UnlockSurface(s);}


int IsTeletext=0;


/* Bit assignments in control reg:
   0 - Flash colour (0=first colour, 1=second)
   1 - Teletext select (0=on chip serialiser, 1=teletext)
 2,3 - Bytes per line (2,3=1,1 is 80, 1,0=40, 0,1=20, 0,0=10)
   4 - CRTC Clock chip select (0 = low frequency, 1= high frequency)
 5,6 - Cursor width in bytes (0,0 = 1 byte, 0,1=not defined, 1,0=2, 1,1=4)
   7 - Master cursor width (if set causes large cursor)
*/
EightUChars FastTable[256];
SixteenUChars FastTableDWidth[256]; /* For mode 4,5,6 */

NarrowResolution FastTableSDL[256];
WideResolution FastTableDWidthSDL[256]; /* For modes 4, 5 and 6 */

int FastTable_Valid=0;

typedef void (*LineRoutinePtr)(void);
LineRoutinePtr LineRoutine;

/* Translates middle bits of VideoULA_ControlReg to number of colours */
static int NColsLookup[]={16, 4, 2, 0 /* Not supported 16? */, 0 /* ???? */, 16, 4, 2}; /* Based on AUG 379 */

unsigned char VideoULA_ControlReg=0x9c;
unsigned char VideoULA_Palette[16];

unsigned char CRTCControlReg=0;
unsigned char CRTC_HorizontalTotal=127;     /* R0 */
unsigned char CRTC_HorizontalDisplayed=80;  /* R1 */
unsigned char CRTC_HorizontalSyncPos=98;    /* R2 */
unsigned char CRTC_SyncWidth=0x28;          /* R3 - top 4 bits are Vertical (in scan lines) and bottom 4 are horizontal in characters */
unsigned char CRTC_VerticalTotal=38;        /* R4 */
unsigned char CRTC_VerticalTotalAdjust=0;   /* R5 */
unsigned char CRTC_VerticalDisplayed=32;    /* R6 */
unsigned char CRTC_VerticalSyncPos=34;      /* R7 */
unsigned char CRTC_InterlaceAndDelay=0;     /* R8 - 0,1 are interlace modes, 4,5 display blanking delay, 6,7 cursor blanking delay */
unsigned char CRTC_ScanLinesPerChar=7;      /* R9 */
unsigned char CRTC_CursorStart=0;           /* R10 */
unsigned char CRTC_CursorEnd=0;             /* R11 */
unsigned char CRTC_ScreenStartHigh=6;       /* R12 */
unsigned char CRTC_ScreenStartLow=0;        /* R13 */
unsigned char CRTC_CursorPosHigh=0;         /* R14 */
unsigned char CRTC_CursorPosLow=0;          /* R15 */
unsigned char CRTC_LightPenHigh=0;          /* R16 */
unsigned char CRTC_LightPenLow=0;           /* R17 */

unsigned int FrameCount=0;

/* CharLine counts from the 'reference point' - i.e. the point at which we reset the address pointer - NOT
the point of the sync. If it is -ve its actually in the adjust time */
typedef struct {
  int Addr;       /* Address of start of next visible character line in beeb memory  - raw */
  int PixmapLine; /* Current line in the pixmap */
  int PreviousFinalPixmapLine; /* The last pixmap line on the previous frame */
  int IsTeletext; /* This frame is a teletext frame - do things differently */
  char *DataPtr;  /* Pointer into host memory of video data */

  int CharLine;   /* 6845 counts in characters vertically - 0 at the top , incs by 1 - -1 means we are in the bit before the actual display starts */
  int InCharLine; /* Scanline within a character line - counts down*/
  int InCharLineUp; /* Scanline within a character line - counts up*/
  int VSyncState; // Cannot =0 in MSVC $NRM; /* When >0 VSync is high */
} VideoStateT;

static  VideoStateT VideoState;

int VideoTriggerCount=9999; /* Number of cycles before next scanline service */

/* first subscript is graphics flag (1 for graphics,2 for separated graphics), next is character, then scanline */
/* character is (valu &127)-32 */
static unsigned char Mode7Font[3][96][9];

static int Mode7FlashOn=1; /* True if a flashing character in mode 7 is on */
static int Mode7DoubleHeightFlags[80]; /* Pessimistic size for this flags - if 1 then corresponding character on NEXT line is top half */

/* Flash every half second(???) i.e. 25 x 50Hz fields */
#define MODE7FLASHFREQUENCY 25
int Mode7FlashTrigger=MODE7FLASHFREQUENCY;

/* If 1 then refresh on every display, else refresh every n'th display */
int Video_RefreshFrequency=1;
/* The number of the current frame - starts at Video_RefreshFrequency - at 0 actually refresh */
//static int FrameNum=0;
static int FrameNum=0;

int time_slept=0;

unsigned char scanlines_scaled[256] = {
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0xff,
        0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0xff,
        0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0xff,
        0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0xff,
        0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0xff,
        0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0xff,
        0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0xff,
        0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0xff,
        0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0xff,
        0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0xff,
        0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xff,
        0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xff,
        0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xff,
        0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xff,
        0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xff,
        0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xff
};

unsigned char scanlines_top[256] = {
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
        0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
        0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
        0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
        0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
        0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
        0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
        0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
        0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
        0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

unsigned char scanlines_bottom[256] = {
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
        0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
        0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
        0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
        0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
        0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
        0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
        0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
        0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
        0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
};

unsigned char scanlines_center[256] = {
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
        0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
        0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
        0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
        0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
        0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
        0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
        0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
        0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
        0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
        0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
        0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
        0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
        0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
        0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
        0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
};

/*
unsigned char scanlines_native[256] = {
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
        0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
        0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
        0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
        0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
        0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
        0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
        0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
        0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
        0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
        0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
};
*/

unsigned char *scanline_mapping_p = scanlines_scaled;

/* 0 = off, 1 = bottom, 2 = top:
 */
unsigned int virtual_keyboard = 0;

void SetVideoCentered(void)
{
	config.settings.vscale = CENTER;
	scanline_mapping_p = scanlines_center;
}

void SetVideoTop(void)
{
	config.settings.vscale = TOP;
	scanline_mapping_p = scanlines_top;
}

void SetVideoBottom(void)
{
	config.settings.vscale = BOTTOM;
	scanline_mapping_p = scanlines_bottom;
}

void SetVideoScaled(void)
{
	config.settings.vscale = SCALED;
	scanline_mapping_p = scanlines_scaled;
}


static void LowLevelDoScanLineNarrow();
static void LowLevelDoScanLineWide();
static void LowLevelDoScanLineNarrowNot4Bytes();
static void LowLevelDoScanLineWideNot4Bytes();

static void LowLevelDoScanLineNarrowSDL();
static void LowLevelDoScanLineWideSDL();
static void LowLevelDoScanLineNarrowNot4BytesSDL();
static void LowLevelDoScanLineWideNot4BytesSDL();

static void LowLevelDoScanLineNarrowSDL_H();
static void LowLevelDoScanLineWideSDL_H();
static void LowLevelDoScanLineNarrowNot4BytesSDL_H();
static void LowLevelDoScanLineWideNot4BytesSDL_H();

static void LowLevelDoScanLineNarrowSDL_F();
static void LowLevelDoScanLineWideSDL_F();
static void LowLevelDoScanLineNarrowNot4BytesSDL_F();
static void LowLevelDoScanLineWideNot4BytesSDL_F();

static void LowLevelDoScanLineNarrowSDL_HF();
static void LowLevelDoScanLineWideSDL_HF();
static void LowLevelDoScanLineNarrowNot4BytesSDL_HF();
static void LowLevelDoScanLineWideNot4BytesSDL_HF();

static void VideoAddCursor(void);

/*-------------------------------------------------------------------------------------------------------------*/
static void BuildMode7Font(void) {
  int presentchar,presentline,presentpixel,presgraph;
  char *tempptr;
  unsigned char  tempvalue;
  /* cout <<"Building mode 7 font data structures\n"; */

  for(presentchar=0;presentchar<96;presentchar++) {
    for(presentline=0;presentline<9;presentline++) {
      /* We build the value of the pixel up */
      tempvalue=0;
      /* Pointer to left pixel in raw font */
      tempptr=mode7fontRaw[(presentchar/16)*9+presentline]+8*(presentchar % 16);

      /* Now each pixel in that character line */
      for(presentpixel=0;presentpixel<8;presentpixel++) {
        if (tempptr[presentpixel]!=' ') tempvalue|=1<<(7-presentpixel);
      }; /* presentpixel */

      /* Store the data into the non graphic as well as the graphic areas - then we will overwrite the graphics one */
      Mode7Font[2][presentchar][presentline]=Mode7Font[1][presentchar][presentline]=Mode7Font[0][presentchar][presentline]=tempvalue;
    }; /* presentline */
  }; /* presentchar */

  /* fill in all the graphics characters */
  for(presgraph=0;presgraph<64;presgraph++) {
    presentchar=(presgraph & 31) + ((presgraph & 32)?64:0);
    Mode7Font[1][presentchar][0]=Mode7Font[1][presentchar][1]=Mode7Font[1][presentchar][2]=((presgraph &1)?0xf0:0) | ((presgraph &2)?0xf:0);
    Mode7Font[1][presentchar][3]=Mode7Font[1][presentchar][4]=Mode7Font[1][presentchar][5]=((presgraph &4)?0xf0:0) | ((presgraph &8)?0xf:0);
    Mode7Font[1][presentchar][6]=Mode7Font[1][presentchar][7]=Mode7Font[1][presentchar][8]=((presgraph &16)?0xf0:0) | ((presgraph &32)?0xf:0);
    /* The following lines do separated graphics */
    Mode7Font[2][presentchar][2]=Mode7Font[2][presentchar][5]=Mode7Font[2][presentchar][8]=0;
    Mode7Font[2][presentchar][0]=Mode7Font[2][presentchar][1]=((presgraph &1)?0xe0:0) | ((presgraph &2)?0xe:0);
    Mode7Font[2][presentchar][3]=Mode7Font[2][presentchar][4]=((presgraph &4)?0xe0:0) | ((presgraph &8)?0xe:0);
    Mode7Font[2][presentchar][6]=Mode7Font[2][presentchar][7]=((presgraph &16)?0xe0:0) | ((presgraph &32)?0xe:0);
  } /*  presgraph */;
}; /* BuildMode7Font */
/*-------------------------------------------------------------------------------------------------------------*/
static void DoFastTable16(void) {
  unsigned int beebpixvl,beebpixvr;
  unsigned int bplvopen,bplvtotal;
  unsigned char tmp;

  for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
    bplvopen=((beebpixvl & 8)?128:0) |
             ((beebpixvl & 4)?32:0) |
             ((beebpixvl & 2)?8:0) |
             ((beebpixvl & 1)?2:0);
    for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
      bplvtotal=bplvopen |
             ((beebpixvr & 8)?64:0) |
             ((beebpixvr & 4)?16:0) |
             ((beebpixvr & 2)?4:0) |
             ((beebpixvr & 1)?1:0);
      tmp=VideoULA_Palette[beebpixvl];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTable[bplvtotal].data[0]=FastTable[bplvtotal].data[1]=
        FastTable[bplvtotal].data[2]=FastTable[bplvtotal].data[3]=mainWin->cols[tmp];

      tmp=VideoULA_Palette[beebpixvr];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTable[bplvtotal].data[4]=FastTable[bplvtotal].data[5]=
        FastTable[bplvtotal].data[6]=FastTable[bplvtotal].data[7]=mainWin->cols[tmp];
    }; /* beebpixr */
  }; /* beebpixl */
}; /* DoFastTable16 */

static void DoFastTable16SDL(void) {
  unsigned int beebpixvl,beebpixvr;
  unsigned int bplvopen,bplvtotal;
  unsigned char tmp;

  for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
    bplvopen=((beebpixvl & 8)?128:0) |
             ((beebpixvl & 4)?32:0) |
             ((beebpixvl & 2)?8:0) |
             ((beebpixvl & 1)?2:0);
    for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
      bplvtotal=bplvopen |
             ((beebpixvr & 8)?64:0) |
             ((beebpixvr & 4)?16:0) |
             ((beebpixvr & 2)?4:0) |
             ((beebpixvr & 1)?1:0);
      tmp=VideoULA_Palette[beebpixvl];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTableSDL[bplvtotal].data[0]=FastTableSDL[bplvtotal].data[1]=mainWin->cols[tmp];

      tmp=VideoULA_Palette[beebpixvr];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTableSDL[bplvtotal].data[2]=FastTableSDL[bplvtotal].data[3]=mainWin->cols[tmp];

    }; /* beebpixr */
  }; /* beebpixl */
}; /* DoFastTable16 */



/*-------------------------------------------------------------------------------------------------------------*/
static void DoFastTable16XStep8(void) {
  unsigned int beebpixvl,beebpixvr;
  unsigned int bplvopen,bplvtotal;
  unsigned char tmp;

  for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
    bplvopen=((beebpixvl & 8)?128:0) |
             ((beebpixvl & 4)?32:0) |
             ((beebpixvl & 2)?8:0) |
             ((beebpixvl & 1)?2:0);
    for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
      bplvtotal=bplvopen |
             ((beebpixvr & 8)?64:0) |
             ((beebpixvr & 4)?16:0) |
             ((beebpixvr & 2)?4:0) |
             ((beebpixvr & 1)?1:0);
      tmp=VideoULA_Palette[beebpixvl];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTableDWidth[bplvtotal].data[0]=FastTableDWidth[bplvtotal].data[1]=
        FastTableDWidth[bplvtotal].data[2]=FastTableDWidth[bplvtotal].data[3]=
      FastTableDWidth[bplvtotal].data[4]=FastTableDWidth[bplvtotal].data[5]=
        FastTableDWidth[bplvtotal].data[6]=FastTableDWidth[bplvtotal].data[7]=mainWin->cols[tmp];

      tmp=VideoULA_Palette[beebpixvr];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTableDWidth[bplvtotal].data[8]=FastTableDWidth[bplvtotal].data[9]=
        FastTableDWidth[bplvtotal].data[10]=FastTableDWidth[bplvtotal].data[11]=
      FastTableDWidth[bplvtotal].data[12]=FastTableDWidth[bplvtotal].data[13]=
        FastTableDWidth[bplvtotal].data[14]=FastTableDWidth[bplvtotal].data[15]=mainWin->cols[tmp];
    }; /* beebpixr */
  }; /* beebpixl */
}; /* DoFastTable16XStep8 */

static void DoFastTable16XStep8SDL(void) {
  unsigned int beebpixvl,beebpixvr;
  unsigned int bplvopen,bplvtotal;
  unsigned char tmp;

  for(beebpixvl=0;beebpixvl<16;beebpixvl++) {
    bplvopen=((beebpixvl & 8)?128:0) |
             ((beebpixvl & 4)?32:0) |
             ((beebpixvl & 2)?8:0) |
             ((beebpixvl & 1)?2:0);
    for(beebpixvr=0;beebpixvr<16;beebpixvr++) {
      bplvtotal=bplvopen |
             ((beebpixvr & 8)?64:0) |
             ((beebpixvr & 4)?16:0) |
             ((beebpixvr & 2)?4:0) |
             ((beebpixvr & 1)?1:0);
      tmp=VideoULA_Palette[beebpixvl];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
        FastTableDWidthSDL[bplvtotal].data[0]=FastTableDWidthSDL[bplvtotal].data[1]=
         FastTableDWidthSDL[bplvtotal].data[2]=FastTableDWidthSDL[bplvtotal].data[3]=mainWin->cols[tmp];

      tmp=VideoULA_Palette[beebpixvr];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
        FastTableDWidthSDL[bplvtotal].data[4]=FastTableDWidthSDL[bplvtotal].data[5]=
         FastTableDWidthSDL[bplvtotal].data[6]=FastTableDWidthSDL[bplvtotal].data[7]=mainWin->cols[tmp];

    }; /* beebpixr */
  }; /* beebpixl */
}; /* DoFastTable16XStep8 */



/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses bits 7,5,3,1 for the       */
/* palette address, the next uses 6,4,2,0, the next uses 5,3,1,H (H=High), then 5,2,0,H                        */
static void DoFastTable4(void) {
  unsigned char tmp;
  unsigned long beebpixv,pentry;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    pentry=((beebpixv & 128)?8:0)
           | ((beebpixv & 32)?4:0)
           | ((beebpixv & 8)?2:0)
           | ((beebpixv & 2)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTable[beebpixv].data[0]=FastTable[beebpixv].data[1]=mainWin->cols[tmp];

    pentry=((beebpixv & 64)?8:0)
           | ((beebpixv & 16)?4:0)
           | ((beebpixv & 4)?2:0)
           | ((beebpixv & 1)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTable[beebpixv].data[2]=FastTable[beebpixv].data[3]=mainWin->cols[tmp];

    pentry=((beebpixv & 32)?8:0)
           | ((beebpixv & 8)?4:0)
           | ((beebpixv & 2)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTable[beebpixv].data[4]=FastTable[beebpixv].data[5]=mainWin->cols[tmp];
    pentry=((beebpixv & 16)?8:0)
           | ((beebpixv & 4)?4:0)
           | ((beebpixv & 1)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTable[beebpixv].data[6]=FastTable[beebpixv].data[7]=mainWin->cols[tmp];
  }; /* beebpixv */
}; /* DoFastTable4 */

static void DoFastTable4SDL(void) {
  unsigned char tmp;
  unsigned long beebpixv,pentry;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    pentry=((beebpixv & 128)?8:0)
           | ((beebpixv & 32)?4:0)
           | ((beebpixv & 8)?2:0)
           | ((beebpixv & 2)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableSDL[beebpixv].data[0]=mainWin->cols[tmp];

    pentry=((beebpixv & 64)?8:0)
           | ((beebpixv & 16)?4:0)
           | ((beebpixv & 4)?2:0)
           | ((beebpixv & 1)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableSDL[beebpixv].data[1]=mainWin->cols[tmp];

    pentry=((beebpixv & 32)?8:0)
           | ((beebpixv & 8)?4:0)
           | ((beebpixv & 2)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableSDL[beebpixv].data[2]=mainWin->cols[tmp];
    pentry=((beebpixv & 16)?8:0)
           | ((beebpixv & 4)?4:0)
           | ((beebpixv & 1)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableSDL[beebpixv].data[3]=mainWin->cols[tmp];
  }; /* beebpixv */
}; /* DoFastTable4 */


/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses bits 7,5,3,1 for the       */
/* palette address, the next uses 6,4,2,0, the next uses 5,3,1,H (H=High), then 5,2,0,H                        */
static void DoFastTable4XStep4(void) {
  unsigned char tmp;
  unsigned long beebpixv,pentry;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    pentry=((beebpixv & 128)?8:0)
           | ((beebpixv & 32)?4:0)
           | ((beebpixv & 8)?2:0)
           | ((beebpixv & 2)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTableDWidth[beebpixv].data[0]=FastTableDWidth[beebpixv].data[1]=
    FastTableDWidth[beebpixv].data[2]=FastTableDWidth[beebpixv].data[3]=mainWin->cols[tmp];

    pentry=((beebpixv & 64)?8:0)
           | ((beebpixv & 16)?4:0)
           | ((beebpixv & 4)?2:0)
           | ((beebpixv & 1)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTableDWidth[beebpixv].data[4]=FastTableDWidth[beebpixv].data[5]=
    FastTableDWidth[beebpixv].data[6]=FastTableDWidth[beebpixv].data[7]=mainWin->cols[tmp];

    pentry=((beebpixv & 32)?8:0)
           | ((beebpixv & 8)?4:0)
           | ((beebpixv & 2)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTableDWidth[beebpixv].data[8]=FastTableDWidth[beebpixv].data[9]=
    FastTableDWidth[beebpixv].data[10]=FastTableDWidth[beebpixv].data[11]=mainWin->cols[tmp];
    pentry=((beebpixv & 16)?8:0)
           | ((beebpixv & 4)?4:0)
           | ((beebpixv & 1)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
    FastTableDWidth[beebpixv].data[12]=FastTableDWidth[beebpixv].data[13]=
    FastTableDWidth[beebpixv].data[14]=FastTableDWidth[beebpixv].data[15]=mainWin->cols[tmp];
  }; /* beebpixv */
}; /* DoFastTable4XStep4 */


static void DoFastTable4XStep4SDL(void) {
  unsigned char tmp;
  unsigned long beebpixv,pentry;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    pentry=((beebpixv & 128)?8:0)
           | ((beebpixv & 32)?4:0)
           | ((beebpixv & 8)?2:0)
           | ((beebpixv & 2)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableDWidthSDL[beebpixv].data[0]=FastTableDWidthSDL[beebpixv].data[1]=mainWin->cols[tmp];

    pentry=((beebpixv & 64)?8:0)
           | ((beebpixv & 16)?4:0)
           | ((beebpixv & 4)?2:0)
           | ((beebpixv & 1)?1:0);
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableDWidthSDL[beebpixv].data[2]=FastTableDWidthSDL[beebpixv].data[3]=mainWin->cols[tmp];

    pentry=((beebpixv & 32)?8:0)
           | ((beebpixv & 8)?4:0)
           | ((beebpixv & 2)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableDWidthSDL[beebpixv].data[4]=FastTableDWidthSDL[beebpixv].data[5]=mainWin->cols[tmp];
    pentry=((beebpixv & 16)?8:0)
           | ((beebpixv & 4)?4:0)
           | ((beebpixv & 1)?2:0)
           | 1;
    tmp=VideoULA_Palette[pentry];
    if (tmp>7) {
      tmp&=7;
      if (VideoULA_ControlReg & 1) tmp^=7;
    };
        FastTableDWidthSDL[beebpixv].data[6]=FastTableDWidthSDL[beebpixv].data[7]=mainWin->cols[tmp];

  }; /* beebpixv */
}; /* DoFastTable4XStep4 */


/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses the same pattern as mode 1 */
/* all the way upto the 5th pixel which uses 31hh then 20hh and hten 1hhh then 0hhhh                           */
static void DoFastTable2(void) {
  unsigned char tmp;
  unsigned long beebpixv,beebpixvt,pentry;
  int pix;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    beebpixvt=beebpixv;
    for(pix=0;pix<8;pix++) {
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt<<=1;
      beebpixvt|=1;
      tmp=VideoULA_Palette[pentry];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTable[beebpixv].data[pix]=mainWin->cols[tmp];
    }; /* pix */
  }; /* beebpixv */
}; /* DoFastTable2 */

static void DoFastTable2SDL(void) {
  unsigned char tmp;
  unsigned long beebpixv,beebpixvt,pentry;
  int pix;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    beebpixvt=beebpixv;
    for(pix=0;pix<8;pix++) {
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt<<=1;
      beebpixvt|=1;
      tmp=VideoULA_Palette[pentry];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
        FastTableSDL[beebpixv].data[pix/2]=mainWin->cols[tmp];
    }; /* pix */
  }; /* beebpixv */
}; /* DoFastTable2 */


/*-------------------------------------------------------------------------------------------------------------*/
/* Some guess work and experimentation has determined that the left most pixel uses the same pattern as mode 1 */
/* all the way upto the 5th pixel which uses 31hh then 20hh and hten 1hhh then 0hhhh                           */
static void DoFastTable2XStep2(void) {
  unsigned char tmp;
  unsigned long beebpixv,beebpixvt,pentry;
  int pix;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    beebpixvt=beebpixv;
    for(pix=0;pix<8;pix++) {
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt<<=1;
      beebpixvt|=1;
      tmp=VideoULA_Palette[pentry];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
      FastTableDWidth[beebpixv].data[pix*2]=FastTableDWidth[beebpixv].data[pix*2+1]=mainWin->cols[tmp];
    }; /* pix */
  }; /* beebpixv */
}; /* DoFastTable2XStep2 */

static void DoFastTable2XStep2SDL(void) {
  unsigned char tmp;
  unsigned long beebpixv,beebpixvt,pentry;
  int pix;

  for(beebpixv=0;beebpixv<256;beebpixv++) {
    beebpixvt=beebpixv;
    for(pix=0;pix<8;pix++) {
      pentry=((beebpixvt & 128)?8:0)
             | ((beebpixvt & 32)?4:0)
             | ((beebpixvt & 8)?2:0)
             | ((beebpixvt & 2)?1:0);
      beebpixvt<<=1;
      beebpixvt|=1;
      tmp=VideoULA_Palette[pentry];
      if (tmp>7) {
        tmp&=7;
        if (VideoULA_ControlReg & 1) tmp^=7;
      };
        FastTableDWidthSDL[beebpixv].data[pix]=mainWin->cols[tmp];
    }; /* pix */
  }; /* beebpixv */
}; /* DoFastTable2XStep2 */


/*-------------------------------------------------------------------------------------------------------------*/
/* Check validity of fast table, and if invalid rebuild.
   The fast table accelerates the translation of beeb video memory
   values into X pixel values */
static void DoFastTable(void) {
  /* if it's already OK then quit */
  if (FastTable_Valid) return;

  if (!(CRTC_HorizontalDisplayed & 3)) {

//	if (mainWin->OnScreenKeyboardShown()) {
//		if (mainWin->ScreenIsReversed()) {
//			// Half width with screen reversed.
//			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowSDL_HF:LowLevelDoScanLineWideSDL_HF;
//		} else {
//			// Half width with normal screen.
//			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowSDL_H:LowLevelDoScanLineWideSDL_H;
//		}
//	} else {
		if (mainWin->ScreenIsReversed()) {
			// Full width with screen reversed.
			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowSDL_F:LowLevelDoScanLineWideSDL_F;
		} else {
			// Full width with normal screen.
			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowSDL:LowLevelDoScanLineWideSDL;
		}
//	}

//    LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowSDL_HF:LowLevelDoScanLineWideSDL_HF;
  } else {

//	if (mainWin->OnScreenKeyboardShown()) {
//		if (mainWin->ScreenIsReversed()) {
//			// Half width with screen reversed.
//			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowNot4BytesSDL_HF:LowLevelDoScanLineWideNot4BytesSDL_HF;
//		} else {
//			// Half width with normal screen.
//			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowNot4BytesSDL_H:LowLevelDoScanLineWideNot4BytesSDL_H;
//		}
//	} else {
		if (mainWin->ScreenIsReversed()) {
			// Full width with screen reversed.
			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowNot4BytesSDL_F:LowLevelDoScanLineWideNot4BytesSDL_F;
		} else {
			// Full width with normal screen.
			LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowNot4BytesSDL:LowLevelDoScanLineWideNot4BytesSDL;
		}
//	}

//    LineRoutine=(VideoULA_ControlReg & 0x10)?LowLevelDoScanLineNarrowNot4BytesSDL_HF:LowLevelDoScanLineWideNot4BytesSDL_HF;
  };

  /* What happens next dpends on the number of colours */
  switch (NColsLookup[(VideoULA_ControlReg & 0x1c) >> 2]) {
    case 2:
      if (VideoULA_ControlReg & 0x10) {
        DoFastTable2SDL();
      } else {
        DoFastTable2XStep2SDL();
      };
      FastTable_Valid=1;     
      break;

    case 4:
      if (VideoULA_ControlReg & 0x10) {
        DoFastTable4SDL();
      } else {
        DoFastTable4XStep4SDL();
      };
      FastTable_Valid=1;     
      break;

    case 16:
      if (VideoULA_ControlReg & 0x10) {
        DoFastTable16SDL();
      } else {
        DoFastTable16XStep8SDL();
      };
      FastTable_Valid=1;
      break;

    default:
      break;
  }; /* Colours/pixel switch */
}; /* DoFastTable */


/*-------------------------------------------------------------------------------------------------------------*/
#define BEEB_DOTIME_SAMPLESIZE 50

static void VideoStartOfFrame(void) {
  int tmp;
  static int InterlaceFrame=0;

	static unsigned int last_time=0;
	unsigned int current_time;
 	int adjusted_time;

//==  static int Have_GotTime=0;
//==  static struct tms previous,now;
//==  static int Time_FrameCount=0;

//==  double frametime;
//==  static CycleCountT OldCycles=0;


//==  if (!Have_GotTime) {
//==    times(&previous);
//==    Time_FrameCount=-1;
//==    Have_GotTime=1;
//==  };

//==  if (Time_FrameCount==(BEEB_DOTIME_SAMPLESIZE-1)) {
//==    times(&now);
//==    frametime=now.tms_utime-previous.tms_utime;
//==    frametime/=(double)CLOCKS_PER_SEC;
//==
//==    OldCycles=TotalCycles;
//==    previous=now;
//==    Time_FrameCount=0;
//==  } else Time_FrameCount++;



	/* Inc frame count and every odd frame synchronize simulated Beeb time with the
	 * Host's actual time. 
	 *
	 * Timing:
	 *
	 * 1000ms / 25fps = 40ms of real time should pass between video repaints.
	 */

#define SYNC 40

	if ( (++FrameCount&1) == 1 && mainWin != NULL) {
		unsigned int current_time = mainWin->GetTicks();
		adjusted_time = SYNC - ( (int) ( current_time - last_time) );
		if (adjusted_time > SYNC) adjusted_time = SYNC;

		if (adjusted_time > 0 && adjusted_time <= SYNC) {
			time_slept += adjusted_time;
			mainWin->Sleep(adjusted_time);
		}

		last_time = mainWin->GetTicks();
	}


  /* If FrameNum hits 0 we actually refresh */
  if (FrameNum--==0) {
    FrameNum=Video_RefreshFrequency-1;
  };

  if (CRTC_VerticalTotalAdjust==0) {
    VideoState.CharLine=0;
    VideoState.InCharLine=CRTC_ScanLinesPerChar;
    VideoState.InCharLineUp=0;
  } else {
    VideoState.CharLine=-1;
    VideoState.InCharLine=CRTC_VerticalTotalAdjust;
    VideoState.InCharLineUp=0;
  };
  VideoState.IsTeletext=(VideoULA_ControlReg &2)>0;
  if (!VideoState.IsTeletext) {
    VideoState.Addr=CRTC_ScreenStartLow+(CRTC_ScreenStartHigh<<8);
  } else {
    int tmphigh=CRTC_ScreenStartHigh;
    /* undo wrangling of start address - I don't understand why this should be - see p.372 of AUG for this info */
    tmphigh^=0x20;
    tmphigh+=0x74;
    tmphigh&=255;
    VideoState.Addr=CRTC_ScreenStartLow+(tmphigh<<8);

    Mode7FlashTrigger--;
    if (Mode7FlashTrigger<0) {
      Mode7FlashTrigger=MODE7FLASHFREQUENCY;
      Mode7FlashOn^=1; /* toggle flash state */
    };
    for(tmp=0;tmp<80;tmp++) {
      Mode7DoubleHeightFlags[tmp]=1; /* corresponding character on NEXT line is top half */
    };
  };

  InterlaceFrame^=1;
  if (InterlaceFrame) {
    IncTrigger((2*(CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2)),VideoTriggerCount); /* Number of 2MHz cycles until another scanline needs doing */
  } else {
    IncTrigger(((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2)),VideoTriggerCount); /* Number of 2MHz cycles until another scanline needs doing */
  };

//	if (mainWin->OnScreenKeyboardShown())
//		VirtualKeyboard_UpdateLEDs();


}; /* VideoStartOfFrame */


/*--------------------------------------------------------------------------*/
/* Scanline processing for modes with fast 6845 clock - i.e. narrow pixels  */
static void LowLevelDoScanLineNarrow() {
  unsigned char *CurrentPtr;
  int BytesToGo=CRTC_HorizontalDisplayed;
  EightUChars *vidPtr=mainWin->GetLinePtr(VideoState.PixmapLine);
//  EightUChars *vid2Ptr=mainWin->GetLine2Ptr(VideoState.PixmapLine);

  /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
  CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

  /* This should help the compiler - it doesn't need to test for end of loop
     except every 4 entries */
  BytesToGo/=4;
  for(;BytesToGo;CurrentPtr+=32,BytesToGo--) {
//    *(vidPtr++)=*(vid2Ptr++)=FastTable[*CurrentPtr];
//    *(vidPtr++)=*(vid2Ptr++)=FastTable[*(CurrentPtr+8)];
//    *(vidPtr++)=*(vid2Ptr++)=FastTable[*(CurrentPtr+16)];
//    *(vidPtr++)=*(vid2Ptr++)=FastTable[*(CurrentPtr+24)];

	*(vidPtr++)=FastTable[*CurrentPtr];
	*(vidPtr++)=FastTable[*(CurrentPtr+8)];
	*(vidPtr++)=FastTable[*(CurrentPtr+16)];
	*(vidPtr++)=FastTable[*(CurrentPtr+24)];

  };
}; /* LowLevelDoScanLineNarrow() */

static void LowLevelDoScanLineNarrowSDL() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr; //, *vidPtr;
	NarrowResolution *vidPtr;

	if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;


//      vidPtr = GetScanlinePtr( scanline );
//	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) {
                printf("LowLevelDoScanLineNarrow: Scanline too wide...\n");
                BytesToGo=80;
        }

        // [TODO] Center image.
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

        CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
        for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
		*(vidPtr++)=FastTableSDL[*CurrentPtr];

//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[0];
//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[1];
//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[2];
//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[3];
        }

//      for(i=0; i<r_border; i++)
//              *(vidPtr2++)=0;

};

static void LowLevelDoScanLineNarrowSDL_H() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;

	if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) {
                printf("LowLevelDoScanLineNarrow: Scanline too wide...\n");
                BytesToGo=80;
        }

        // [TODO] Center image.
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

        CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
        for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableSDL[*CurrentPtr];

                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[0];
//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[1];
                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[2];
//                *(vidPtr++)=FastTableSDL[*CurrentPtr].data[3];
        }

//      for(i=0; i<r_border; i++)
//              *(vidPtr2++)=0;

};

static void LowLevelDoScanLineNarrowSDL_F() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;

	if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) {
                printf("LowLevelDoScanLineNarrow: Scanline too wide...\n");
                BytesToGo=80;
        }

        // [TODO] Center image.
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

	vidPtr+=(BytesToGo<<2)-1;

        CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
        for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableSDL[*CurrentPtr];

                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[0];
                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[1];
                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[2];
                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[3];
        }

//      for(i=0; i<r_border; i++)
//              *(vidPtr2++)=0;

};

static void LowLevelDoScanLineNarrowSDL_HF() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;

	if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) {
                printf("LowLevelDoScanLineNarrow: Scanline too wide...\n");
                BytesToGo=80;
        }

        // [TODO] Center image.
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

	vidPtr+=(BytesToGo<<2)-1;

        CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
        for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableSDL[*CurrentPtr];

                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[0];
//                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[1];
                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[2];
//                *(vidPtr--)=FastTableSDL[*CurrentPtr].data[3];
        }

//      for(i=0; i<r_border; i++)
//              *(vidPtr2++)=0;

};




/*--------------------------------------------------------------------------*/
/* Scanline processing for modes with fast 6845 clock - i.e. narrow pixels  */
/* This version handles screen modes where there is not a multiple of 4     */
/* bytes per scanline.                                                      */
static void LowLevelDoScanLineNarrowNot4Bytes() {
  unsigned char *CurrentPtr;
  int BytesToGo=CRTC_HorizontalDisplayed;
  EightUChars *vidPtr=mainWin->GetLinePtr(VideoState.PixmapLine);
//  EightUChars *vid2Ptr=mainWin->GetLine2Ptr(VideoState.PixmapLine);

  /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

  for(;BytesToGo;CurrentPtr+=8,BytesToGo--)
//    (vidPtr++)->eightbyte=(vid2Ptr++)->eightbyte=FastTable[*CurrentPtr].eightbyte;
	(vidPtr++)->eightbyte=FastTable[*CurrentPtr].eightbyte;

}; /* LowLevelDoScanLineNarrowNot4Bytes() */

static void LowLevelDoScanLineNarrowNot4BytesSDL() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr; //, *vidPtr;
	NarrowResolution *vidPtr;
        if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine & 0xff];

        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
//      vidPtr = GetScanlinePtr( scanline );
//	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) BytesToGo=80;

        // [TODO]
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
	*(vidPtr++)=FastTableSDL[*CurrentPtr];

//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
}

static void LowLevelDoScanLineNarrowNot4BytesSDL_H() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;
        if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine & 0xff];

        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) BytesToGo=80;

        // [TODO]
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//	*(vidPtr++)=FastTableSDL[*CurrentPtr];

        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
}

static void LowLevelDoScanLineNarrowNot4BytesSDL_F() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;
        if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine & 0xff];

        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) BytesToGo=80;

        // [TODO]
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

	vidPtr+=(BytesToGo<<2)-1;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//	*(vidPtr++)=FastTableSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
}

static void LowLevelDoScanLineNarrowNot4BytesSDL_HF() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	NarrowResolution *vidPtr;
        if (VideoState.PixmapLine>0xff) return;

        scanline = scanline_mapping_p[VideoState.PixmapLine & 0xff];

        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

//      vidPtr = GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (NarrowResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>80) BytesToGo=80;

        // [TODO]
        l_border = 0;
        r_border = 320 - ( (BytesToGo*4) + l_border);

//      for(i=0; i<l_border; i++)
//              *(vidPtr2++)=0;

	vidPtr+=(BytesToGo<<2)-1;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//	*(vidPtr++)=FastTableSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
}




/*-----------------------------------------------------------------------------*/
/* Scanline processing for the low clock rate modes                            */
static void LowLevelDoScanLineWide() {
  unsigned char *CurrentPtr;
  int BytesToGo=CRTC_HorizontalDisplayed;
  SixteenUChars *vidPtr=mainWin->GetLinePtr16(VideoState.PixmapLine);
//  SixteenUChars *vid2Ptr=mainWin->GetLine2Ptr16(VideoState.PixmapLine);

  /* If the step is 4 then each byte corresponds to one entry in the fasttable
     and thus we can copy it really easily (and fast!) */
    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

  /* This should help the compiler - it doesn't need to test for end of loop
     except every 4 entries */
  BytesToGo/=4;
  for(;BytesToGo;CurrentPtr+=32,BytesToGo--) {
//    *(vidPtr++)=*(vid2Ptr++)=FastTableDWidth[*CurrentPtr];
//    *(vidPtr++)=*(vid2Ptr++)=FastTableDWidth[*(CurrentPtr+8)];
//    *(vidPtr++)=*(vid2Ptr++)=FastTableDWidth[*(CurrentPtr+16)];
//    *(vidPtr++)=*(vid2Ptr++)=FastTableDWidth[*(CurrentPtr+24)];

	*(vidPtr++)=FastTableDWidth[*CurrentPtr];
	*(vidPtr++)=FastTableDWidth[*(CurrentPtr+8)];
	*(vidPtr++)=FastTableDWidth[*(CurrentPtr+16)];
	*(vidPtr++)=FastTableDWidth[*(CurrentPtr+24)];

  };
}; /* LowLevelDoScanLineWide */

static void LowLevelDoScanLineWideSDL() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr;
        //SixteenUChars *vidPtr;
	//==unsigned char *vidPtr;
	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
        //vidPtr = GetScanlinePtr( scanline );
        //vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	//==vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[4];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[5];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[6];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
};

static void LowLevelDoScanLineWideSDL_H() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
        //SixteenUChars *vidPtr;
	//==unsigned char *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
        //vidPtr = GetScanlinePtr( scanline );
        //vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[4];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[6];
        //==*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
};

static void LowLevelDoScanLineWideSDL_F() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
        //SixteenUChars *vidPtr;
	//==unsigned char *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

        //vidPtr = GetScanlinePtr( scanline );
        //vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

	vidPtr+=(BytesToGo<<3)-1;

    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[4];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[6];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
};

static void LowLevelDoScanLineWideSDL_HF() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
        //SixteenUChars *vidPtr;
	//==unsigned char *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

        //vidPtr = GetScanlinePtr( scanline );
        //vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr2++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

	vidPtr+=(BytesToGo<<3)-1;

    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[4];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[6];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }

//        for(i=0; i<r_border; i++)
//                *(vidPtr2++)=0;
};




/*-----------------------------------------------------------------------------*/
/* Scanline processing for the low clock rate modes                            */
/* This version handles cases where the screen width is not divisible by 4     */
static void LowLevelDoScanLineWideNot4Bytes() {
  unsigned char *CurrentPtr;
  int BytesToGo=CRTC_HorizontalDisplayed;
  SixteenUChars *vidPtr=mainWin->GetLinePtr16(VideoState.PixmapLine);
//  SixteenUChars *vid2Ptr=mainWin->GetLine2Ptr16(VideoState.PixmapLine);

  CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;

  for(;BytesToGo;CurrentPtr+=8,BytesToGo--)
//    *(vidPtr++)=*(vid2Ptr++)=FastTableDWidth[*CurrentPtr];

  	*(vidPtr++)=FastTableDWidth[*CurrentPtr];	

}; /* LowLevelDoScanLineWideNot4Bytes */

static void LowLevelDoScanLineWideNot4BytesSDL() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr; //, *vidPtr;
	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
	//vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	//vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
		*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[4];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[5];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[6];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }


//        for(i=0; i<r_border; i++)
//                *(vidPtr++)=0;
};

static void LowLevelDoScanLineWideNot4BytesSDL_H() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;
	//vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr++)=0;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[4];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[6];
//        *(vidPtr++)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }


//        for(i=0; i<r_border; i++)
//                *(vidPtr++)=0;
};

static void LowLevelDoScanLineWideNot4BytesSDL_F() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

	//vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr++)=0;

	vidPtr+=(BytesToGo<<3)-1;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[4];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[6];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }


//        for(i=0; i<r_border; i++)
//                *(vidPtr++)=0;
};

static void LowLevelDoScanLineWideNot4BytesSDL_HF() {
        unsigned int scanline, BytesToGo, i, r_border=0, l_border=0;
        unsigned char *CurrentPtr, *vidPtr;
//	WideResolution *vidPtr;

        if (VideoState.PixmapLine>0xff) return;
        scanline = scanline_mapping_p[VideoState.PixmapLine &0xff];
        if (scanline==0xff || ( mainWin->OnScreenKeyboardShown() && scanline >= 105) ) return;

	scanline=abs(239-scanline);

	//vidPtr = (SixteenUChars*) GetScanlinePtr( scanline );
	vidPtr = (unsigned char*) mainWin->GetLinePtr(scanline);
//	vidPtr = (WideResolution*) mainWin->GetLinePtr(scanline);

        BytesToGo=CRTC_HorizontalDisplayed;

        // Sometimes CRTC_HorizontalDisplayed is greater than a scanlines maximum pitch, so clip at 80.
        if (BytesToGo>40) BytesToGo=40;

        l_border = 0;
        r_border = 320 - ( (BytesToGo*8) + l_border);

//        for(i=0; i<l_border; i++)
//                *(vidPtr++)=0;

	vidPtr+=(BytesToGo<<3)-1;

    CurrentPtr=(unsigned char *)VideoState.DataPtr+VideoState.InCharLineUp;
    for(;BytesToGo;CurrentPtr+=8,BytesToGo--) {
//		*(vidPtr++)=FastTableDWidthSDL[*CurrentPtr];

        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[0];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[1];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[2];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[3];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[4];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[5];
        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[6];
//        *(vidPtr--)=FastTableDWidthSDL[*CurrentPtr].data[7];
    }


//        for(i=0; i<r_border; i++)
//                *(vidPtr++)=0;
};



/*-------------------------------------------------------------------------------------------------------------*/
/* Do all the pixel rows for one row of teletext characters                                                    */

// [Here]

static void DoMode7Row(void) {
  char *CurrentPtr=VideoState.DataPtr;
  int CurrentChar;
  int XStep;
  unsigned char byte,tmp;

  unsigned int Foreground=mainWin->cols[7];
  unsigned int ActualForeground;
  unsigned int Background=mainWin->cols[0];
  int Flash=0; /* i.e. steady */
  int DoubleHeight=0; /* Normal */
  int Graphics=0; /* I.e. alpha */
  int Separated=0; /* i.e. continuous graphics */
  int HoldGraph=0; /* I.e. don't hold graphics - I don't know what hold graphics is anyway! */

  unsigned int CurrentCol[9]={0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xffffff};
  int CurrentLen[9]={0,0,0,0,0,0,0,0,0};
  int CurrentStartX[9]={0,0,0,0,0,0,0,0,0};
  int CurrentScanLine;
  int CurrentX=0;
  int CurrentPixel;
  int col;
  int FontTypeIndex=0; /* 0=alpha, 1=continuous graphics, 2=separated graphics */

  unsigned int scanline=0;

  if (CRTC_HorizontalDisplayed>80) return; /* Not possible on beeb - and would break the double height lookup array */

//->  if (CRTC_HorizontalDisplayed>0)
//--    XStep=640/(CRTC_HorizontalDisplayed*8);
//--  else
//--    XStep=1;
//++
XStep=1;
//<-

  for(CurrentChar=0;CurrentChar<CRTC_HorizontalDisplayed;CurrentChar++) {
    byte=CurrentPtr[CurrentChar];
    if ((byte>=128) && (byte<=159)) {
      switch (byte) {
        case 129:
        case 130:
        case 131:
        case 132:
        case 133:
        case 134:
        case 135:
          Foreground=mainWin->cols[byte-128];
          Graphics=0;
          break;

        case 136:
          Flash=1;
          break;

        case 137:
          Flash=0;
          break;

        case 140:
          DoubleHeight=0;
          break;

        case 141:
          DoubleHeight=1;
          break;

        case 145:
        case 146:
        case 147:
        case 148:
        case 149:
        case 150:
        case 151:
          Foreground=mainWin->cols[byte-144];
          Graphics=1;
          break;

        case 152: /* Conceal display - not sure about this */
          Foreground=Background;
          break;

        case 153:
          Separated=0;
          break;

        case 154:
          Separated=1;
          break;

        case 156:
          Background=mainWin->cols[0];
          break;

        case 157:
          Background=Foreground;
          break;

        case 158:
          HoldGraph=1;
          break;

        case 159:
          HoldGraph=0;
          break;
      }; /* Special character switch */

      /* Fudge so that the special character is just displayed in background */
      byte=32;
      FontTypeIndex=Graphics?(Separated?2:1):0;
    }; /* test for special character */

    /* Top bit never reaches character generator */
    byte&=127;
    /* Our font table goes from character 32 up */
    if (byte<32) byte=0; else byte-=32;

    /* Conceal flashed text if necessary */
    ActualForeground=(Flash && !Mode7FlashOn)?Background:Foreground;
    if (!DoubleHeight) {
      for(CurrentScanLine=0;CurrentScanLine<9;CurrentScanLine++) {
        tmp=Mode7Font[FontTypeIndex][byte][CurrentScanLine];
        if ((tmp==0) || (tmp==255)) {
          col=(tmp==0)?Background:ActualForeground;
          if (col==(int) CurrentCol[CurrentScanLine]) CurrentLen[CurrentScanLine]+=8*XStep; else {
            if (CurrentLen[CurrentScanLine])


              mainWin->doHorizLine(CurrentCol[CurrentScanLine],VideoState.PixmapLine+CurrentScanLine,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);

//==               scanline = scanline_mapping_p[VideoState.PixmapLine+CurrentScanLine & 0xff];
//==               if (scanline==0xff) break;
//==		mainWin->doHorizLine(CurrentCol[CurrentScanLine],scanline,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);


            CurrentCol[CurrentScanLine]=col;
            CurrentStartX[CurrentScanLine]=CurrentX;
            CurrentLen[CurrentScanLine]=8*XStep;
          }; /* same colour */
        } else {
          for(CurrentPixel=0x80;CurrentPixel;CurrentPixel=CurrentPixel>>1) {
            /* Background or foreground ? */
            col=(tmp & CurrentPixel)?ActualForeground:Background;

            /* Do we need to draw ? */
            if (col==(int) CurrentCol[CurrentScanLine]) CurrentLen[CurrentScanLine]+=XStep; else {
              if (CurrentLen[CurrentScanLine]) 

                mainWin->doHorizLine(CurrentCol[CurrentScanLine],VideoState.PixmapLine+CurrentScanLine,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);

//==               scanline = scanline_mapping_p[VideoState.PixmapLine+CurrentScanLine & 0xff];
//==               if (scanline==0xff) break;
//==               mainWin->doHorizLine(CurrentCol[CurrentScanLine],scanline,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);


              CurrentCol[CurrentScanLine]=col;
              CurrentStartX[CurrentScanLine]=CurrentX;
              CurrentLen[CurrentScanLine]=XStep;
            }; /* Fore/back ground */
            CurrentX+=XStep;
          }; /* Pixel within byte */
          CurrentX-=8*XStep;
        }; /* tmp!=0 */
      }; /* Scanline for */
      CurrentX+=8*XStep;
      Mode7DoubleHeightFlags[CurrentChar]=1; /* Not double height - so if the next line is double height it will be top half */
    } else {
      int ActualScanLine;
      /* Double height! */
      for(CurrentPixel=0x80;CurrentPixel;CurrentPixel=CurrentPixel>>1) {
        for(CurrentScanLine=0;CurrentScanLine<9;CurrentScanLine++) {
          if (Mode7DoubleHeightFlags[CurrentChar]) ActualScanLine=CurrentScanLine >> 1; else ActualScanLine=4+((CurrentScanLine+1)>>1);
          /* Background or foreground ? */
          col=(Mode7Font[FontTypeIndex][byte][ActualScanLine] & CurrentPixel)?ActualForeground:Background;

          /* Do we need to draw ? */
          if (col==(int) CurrentCol[CurrentScanLine]) CurrentLen[CurrentScanLine]+=XStep; else {
            if (CurrentLen[CurrentScanLine])  {

              mainWin->doHorizLine(CurrentCol[CurrentScanLine],VideoState.PixmapLine+CurrentScanLine,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);

//==               scanline = scanline_mapping_p[VideoState.PixmapLine+CurrentScanLine & 0xff];
//==               if (scanline==0xff) break;
//==               mainWin->doHorizLine(CurrentCol[CurrentScanLine],scanline,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);


            };
            CurrentCol[CurrentScanLine]=col;
            CurrentStartX[CurrentScanLine]=CurrentX;
            CurrentLen[CurrentScanLine]=XStep;
          }; /* Fore/back ground */
        }; /* Scanline for */
        CurrentX+=XStep;
      }; /* Pixel within byte */
      Mode7DoubleHeightFlags[CurrentChar]^=1; /* Not double height - so if the next line is double height it will be top half */
    };
  }; /* character loop */

  /* Finish off right bits of scan line */
  for(CurrentScanLine=0;CurrentScanLine<9;CurrentScanLine++) {
    if (CurrentLen[CurrentScanLine])

      mainWin->doHorizLine(CurrentCol[CurrentScanLine],VideoState.PixmapLine+CurrentScanLine,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);

//==	scanline = scanline_mapping_p[VideoState.PixmapLine+CurrentScanLine & 0xff];
//==	if (scanline==0xff) break;
//==	mainWin->doHorizLine(CurrentCol[CurrentScanLine],scanline,CurrentStartX[CurrentScanLine],CurrentLen[CurrentScanLine]);


  };
}; /* DoMode7Row */




/*-------------------------------------------------------------------------------------------------------------*/
/* Actually does the work of decoding beeb memory and plotting the line to X */
static void LowLevelDoScanLine() {
  /* Update acceleration tables */
  DoFastTable();
  if (FastTable_Valid) LineRoutine();
}; /* LowLevelDoScanLine */

/*-------------------------------------------------------------------------------------------------------------*/
void VideoDoScanLine(void) {

	LOCK(frame_buffer_p);

	IsTeletext =VideoState.IsTeletext;


  /* cerr << "CharLine=" << VideoState.CharLine << " InCharLine=" << VideoState.InCharLine << "\n"; */
  if (VideoState.IsTeletext) {
    static int DoCA1Int=0;

    if (DoCA1Int) {
      SysVIATriggerCA1Int(0);
      DoCA1Int=0;
    };

    if ((VideoState.CharLine!=-1) && (VideoState.CharLine<CRTC_VerticalDisplayed)) {
      VideoState.DataPtr=BeebMemPtrWithWrapMo7(VideoState.Addr,CRTC_HorizontalDisplayed);
      VideoState.Addr+=CRTC_HorizontalDisplayed;
      if (!FrameNum) {

		if ((FrameCount&1) == 1)
		DoMode7Row();

	}
      VideoState.PixmapLine+=9;
    };


      /* Move onto next physical scanline as far as timing is concerned */
    if (VideoState.CharLine==-1) {
      VideoState.InCharLine-=1;
      VideoState.InCharLineUp+=1;
    } else 
      VideoState.InCharLine=-1;

      if (VideoState.InCharLine<0) {
        VideoState.CharLine++;
        VideoState.InCharLine=CRTC_ScanLinesPerChar;
        VideoState.InCharLineUp=0;
      };

    if (VideoState.CharLine>CRTC_VerticalTotal) {
      if (!FrameNum) {

//[Here]
        mainWin->updateLines(0,256);
        //VideoAddCursor();

        /* Fill unscanned lines under picture.  Cursor will displayed on one of these
           lines when its on the last line of the screen so they are cleared after they
           are displayed, ready for the next screen. */
        if (VideoState.PixmapLine<256) {

//[Here]
//          memset(mainWin->imageData()+VideoState.PixmapLine*640*2,
//                 mainWin->cols[0], (256-VideoState.PixmapLine)*640*2);
        }
      }
      VideoStartOfFrame();
      VideoState.PreviousFinalPixmapLine=VideoState.PixmapLine;
      VideoState.PixmapLine=0;
      SysVIATriggerCA1Int(1);
      DoCA1Int=1;
    } else {
      if (VideoState.CharLine!=-1)  {
        IncTrigger((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2)*((VideoState.CharLine==1)?9:10),VideoTriggerCount);
      } else {
        IncTrigger((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2),VideoTriggerCount);
      };
    };
  } else {
    /* Non teletext */
    if ((VideoState.CharLine!=-1) && (VideoState.CharLine<CRTC_VerticalDisplayed)) {
      /* If first row of character then get the data pointer from memory */
      if (VideoState.InCharLine==CRTC_ScanLinesPerChar) {
        VideoState.DataPtr=BeebMemPtrWithWrap(VideoState.Addr*8,CRTC_HorizontalDisplayed*8);
        VideoState.Addr+=CRTC_HorizontalDisplayed;
      };

      if ((VideoState.InCharLine)>(CRTC_ScanLinesPerChar-8)) {

        if (!FrameNum) if ( ((mainWin->currentframe())&1)==1) LowLevelDoScanLine();
        VideoState.PixmapLine+=1;

      } //==else {
        //==if (!FrameNum) {}//==mainWin->doHorizLine(mainWin->cols[0],VideoState.PixmapLine++,0,640);
      //==};
    }; /* !=-1 and is in displayed bit */


    /* Move onto next physical scanline as far as timing is concerned */
    VideoState.InCharLine-=1;
    VideoState.InCharLineUp+=1;
    if (VideoState.VSyncState) {
      if (!(--VideoState.VSyncState)) {
        SysVIATriggerCA1Int(0);
      };
    };

    if (VideoState.InCharLine<0) {
      VideoState.CharLine++;
      /* Suspect the -1 in sync pos is a fudge factor - careful!  - DAG - now out */
      if ((VideoState.VSyncState==0) && (VideoState.CharLine==(CRTC_VerticalSyncPos))) {
        /* Fill unscanned lines under picture */
        if (VideoState.PixmapLine<VideoState.PreviousFinalPixmapLine) {
          int CurrentLine;
          for(CurrentLine=VideoState.PixmapLine;(CurrentLine<256) && (CurrentLine<VideoState.PreviousFinalPixmapLine);
              CurrentLine++) {
            mainWin->doHorizLine(mainWin->cols[0],CurrentLine,0,640);
          };
        };
        VideoState.PreviousFinalPixmapLine=VideoState.PixmapLine;
        VideoState.PixmapLine=0;
        SysVIATriggerCA1Int(1);
        VideoState.VSyncState=3;
      };

      VideoState.InCharLine=CRTC_ScanLinesPerChar;
      VideoState.InCharLineUp=0;
    };

    if (VideoState.CharLine>CRTC_VerticalTotal) {
      if (!FrameNum) {

//[Here]

        mainWin->updateLines(0,256);
        //VideoAddCursor();
      }
      VideoStartOfFrame();
    } else
      IncTrigger((CRTC_HorizontalTotal+1)*((VideoULA_ControlReg & 16)?1:2),VideoTriggerCount);
  }; /* Teletext if */

	// ++ added this one
	VideoAddCursor();

	UNLOCK(frame_buffer_p);
}; /* VideoDoScanLine */

/*-------------------------------------------------------------------------------------------------------------*/
void VideoInit(void) {
  char *environptr;
  VideoStartOfFrame();
  VideoState.DataPtr=BeebMemPtrWithWrap(0x3000,640);
  SetTrigger(99,VideoTriggerCount); /* Give time for OS to set mode up before doing anything silly */
  FastTable_Valid=0;
  BuildMode7Font();

#ifndef WIN32
  environptr=getenv("BeebVideoRefreshFreq");
  if (environptr!=NULL) Video_RefreshFrequency=atoi(environptr);
  if (Video_RefreshFrequency<1) Video_RefreshFrequency=1;
#endif

  FrameNum=Video_RefreshFrequency;
  VideoState.PixmapLine=0;
  VideoState.PreviousFinalPixmapLine=255;
}; /* VideoInit */

/*-------------------------------------------------------------------------------------------------------------*/
void CRTCWrite(int Address, int Value) {
  Value&=0xff;
  if (Address & 1) {
    switch (CRTCControlReg) {
      case 0:
        CRTC_HorizontalTotal=Value;
        break;

      case 1:
        CRTC_HorizontalDisplayed=Value;
        FastTable_Valid=0;
        break;

      case 2:
        CRTC_HorizontalSyncPos=Value;
        break;

      case 3:
        CRTC_SyncWidth=Value;
        break;

      case 4:
        CRTC_VerticalTotal=Value;
        break;

      case 5:
        CRTC_VerticalTotalAdjust=Value;
        break;

      case 6:
        CRTC_VerticalDisplayed=Value;
        break;

      case 7:
        CRTC_VerticalSyncPos=Value;
        break;

      case 8:
        CRTC_InterlaceAndDelay=Value;
        break;

      case 9:
        CRTC_ScanLinesPerChar=Value;
        break;

      case 10:
        CRTC_CursorStart=Value;
        break;

      case 11:
        CRTC_CursorEnd=Value;
        break;

      case 12:
        CRTC_ScreenStartHigh=Value;
        break;

      case 13:
        CRTC_ScreenStartLow=Value;
        break;

      case 14:
        CRTC_CursorPosHigh=Value & 0x3f; /* Cursor high only has 6 bits */
        break;

      case 15:
        CRTC_CursorPosLow=Value & 0xff;
        break;

      default: /* In case the user wrote a duff control register value */
        break;
    }; /* CRTCWrite switch */
    /* cerr << "CRTCWrite RegNum=" << int(CRTCControlReg) << " Value=" << Value << "\n"; */
  } else {
    CRTCControlReg=Value & 0x1f;
  };
}; /* CRTCWrite */

/*-------------------------------------------------------------------------------------------------------------*/
int CRTCRead(int Address) {
  if (Address & 1) {
    switch (CRTCControlReg) {
      case 14:
        return(CRTC_CursorPosHigh);
      case 15:
        return(CRTC_CursorPosLow);
      case 16:
        return(CRTC_LightPenHigh); /* Perhaps tie to mouse pointer ? */
      case 17:
        return(CRTC_LightPenLow);
      default:
        break;
    }; /* CRTC Read switch */
  } else {
    return(0); /* Rockwell part has bits 5,6,7 used - bit 6 is set when LPEN is received, bit 5 when in vertical retrace */
  }
return(0);	// Keeep MSVC happy $NRM
}; /* CRTCRead */

/*-------------------------------------------------------------------------------------------------------------*/
void VideoULAWrite(int Address, int Value) {
  if (Address & 1) {
    VideoULA_Palette[(Value & 0xf0)>>4]=(Value & 0xf) ^ 7;
    FastTable_Valid=0;
    /* cerr << "Palette reg " << ((Value & 0xf0)>>4) << " now has value " << ((Value & 0xf) ^ 7) << "\n"; */
  } else {
    VideoULA_ControlReg=Value;
    FastTable_Valid=0; /* Could be more selective and only do it if no.of.cols bit changes */
    /* cerr << "VidULA Ctrl reg write " << hex << Value << "\n"; */
  };
}; /* VidULAWrite */

/*-------------------------------------------------------------------------------------------------------------*/
int VideoULARead(int Address) {
  return(Address); /* Read not defined from Video ULA */
}; /* VidULARead */

/*-------------------------------------------------------------------------------------------------------------*/

#define CURSORFLASHRATE 20

static void VideoAddCursor(void) {
	static int CurSizes[] = { 2,1,0,0,4,2,0,4 };

	static int cursorvisible = 0; //, cursorstate = 0;

	static unsigned int my_frame = 0;


//	VideoWidth= CRTC_HorizontalDisplayed*8;

//	printf ("%d, %d\n", cursorvisible, cursorstate);


	if (config.settings.disable_cursor) return;



	/* Check if cursor has been hidden */
	if ((VideoULA_ControlReg & 0xe0) == 0 || (CRTC_CursorStart & 0x40) == 0)
		return;



	if ( my_frame <= FrameCount ) {
//	if (cursorstate<=0){
//		cursorstate=CURSORFLASHRATE;
		my_frame = FrameCount + CURSORFLASHRATE;

		
		if(cursorvisible == 0)
			cursorvisible=1;
		else
			cursorvisible=0;

//	}else{
//		cursorstate-=1;
	}

		


	int ScrAddr,CurAddr,RelAddr;
	int CurX,CurY;
	int CurSize;
	int CurStart, CurEnd;

	/* Use clock bit and cursor buts to work out size */
	if (VideoULA_ControlReg & 0x80)
		CurSize = CurSizes[(VideoULA_ControlReg & 0x70)>>4] * 8;
	else
		CurSize = 2 * 8; /* Mode 7 */

	if (VideoState.IsTeletext)
	{
		ScrAddr=CRTC_ScreenStartLow+(((CRTC_ScreenStartHigh ^ 0x20) + 0x74 & 0xff)<<8);
		CurAddr=CRTC_CursorPosLow+(((CRTC_CursorPosHigh ^ 0x20) + 0x74 & 0xff)<<8);

		CurStart = (CRTC_CursorStart & 0x1f) / 2;
		CurEnd = CRTC_CursorEnd / 2;
	}
	else
	{
		ScrAddr=CRTC_ScreenStartLow+(CRTC_ScreenStartHigh<<8);
		CurAddr=CRTC_CursorPosLow+(CRTC_CursorPosHigh<<8);

		CurStart = CRTC_CursorStart & 0x1f;
		CurEnd = CRTC_CursorEnd;
	}
		
	RelAddr=CurAddr-ScrAddr;
	if (RelAddr < 0 || CRTC_HorizontalDisplayed == 0)
		return;

	/* Work out char positions */
	CurX = RelAddr % CRTC_HorizontalDisplayed;
	CurY = RelAddr / CRTC_HorizontalDisplayed;

	/* Convert to pixel positions */
	CurX = CurX*640/CRTC_HorizontalDisplayed;

//==	if (CRTC_ScanLinesPerChar != 9)
//==		CurY = CurY * (VideoState.IsTeletext ? CRTC_ScanLinesPerChar/2 : CRTC_ScanLinesPerChar + 1);
//==	else
//==		CurY = CurY * 8;

	CurY*= VideoState.IsTeletext ? 9 : 8;


//	printf("%d\n", CRTC_ScanLinesPerChar);

	/* Limit cursor size */
	if (CurEnd > 9)
		CurEnd = 9;

	if (CurX + CurSize >= 640)
		CurSize = 640 - CurX;

	if (CurSize > 0)
	{
		for (int y = CurStart; y <= CurEnd && CurY + y < 256; ++y)
		{
			if (CurY + y >= 0){


//		printf("CurY=%d, y=%d, CurX=%d, CurSize=%d\n", CurY, y, CurX, CurSize);
	


		if (cursorvisible && Config_DisableCursor == 0)
//				mainWin->doHorizLine(mainWin->cols[7], CurY + (y-1), CurX, CurSize);
			mainWin->doHorizLine(mainWin->cols[7],  ( VideoState.IsTeletext ? CurY + (y-1) : scanline_mapping_p[(CurY + (y-1)) & 0xff] ), CurX>>1, CurSize>>1);


	}



		}
	}
}


/*-------------------------------------------------------------------------*/
void SaveVideoUEF(FILE *SUEF) {
        fput16(0x0468,SUEF);
        fput32(47,SUEF);
        // save the registers now
        fputc(CRTC_HorizontalTotal,SUEF);
        fputc(CRTC_HorizontalDisplayed,SUEF);
        fputc(CRTC_HorizontalSyncPos,SUEF);
        fputc(CRTC_SyncWidth,SUEF);
        fputc(CRTC_VerticalTotal,SUEF);
        fputc(CRTC_VerticalTotalAdjust,SUEF);
        fputc(CRTC_VerticalDisplayed,SUEF);
        fputc(CRTC_VerticalSyncPos,SUEF);
        fputc(CRTC_InterlaceAndDelay,SUEF);
        fputc(CRTC_ScanLinesPerChar,SUEF);
        fputc(CRTC_CursorStart,SUEF);
        fputc(CRTC_CursorEnd,SUEF);
        fputc(CRTC_ScreenStartHigh,SUEF);
        fputc(CRTC_ScreenStartLow,SUEF);
        fputc(CRTC_CursorPosHigh,SUEF);
        fputc(CRTC_CursorPosLow,SUEF);
        fputc(CRTC_LightPenHigh,SUEF);
        fputc(CRTC_LightPenLow,SUEF);
        // VIDPROC
        fputc(VideoULA_ControlReg,SUEF);
        for (int col = 0; col < 16; ++col)
                fputc(VideoULA_Palette[col] ^ 7,SUEF); /* Use real ULA values */

//==        fput16(ActualScreenWidth,SUEF);
	fput16(0, SUEF);

//==        fput32(ScreenAdjust,SUEF);
	fput32(0, SUEF);

        fputc(CRTCControlReg,SUEF);

//==        fputc(TeletextStyle,SUEF);
	fputc(0, SUEF);

        fput32(0,SUEF); // Pad out
}
void LoadVideoUEF(FILE *SUEF) {
	int ActualScreenWidth, ScreenAdjust, TeletextStyle;

        CRTC_HorizontalTotal=fgetc(SUEF);
        CRTC_HorizontalDisplayed=fgetc(SUEF);
        CRTC_HorizontalSyncPos=fgetc(SUEF);
        CRTC_SyncWidth=fgetc(SUEF);
        CRTC_VerticalTotal=fgetc(SUEF);
        CRTC_VerticalTotalAdjust=fgetc(SUEF);
        CRTC_VerticalDisplayed=fgetc(SUEF);
        CRTC_VerticalSyncPos=fgetc(SUEF);
        CRTC_InterlaceAndDelay=fgetc(SUEF);
        CRTC_ScanLinesPerChar=fgetc(SUEF);
        CRTC_CursorStart=fgetc(SUEF);
        CRTC_CursorEnd=fgetc(SUEF);
        CRTC_ScreenStartHigh=fgetc(SUEF);
        CRTC_ScreenStartLow=fgetc(SUEF);
        CRTC_CursorPosHigh=fgetc(SUEF);
        CRTC_CursorPosLow=fgetc(SUEF);
        CRTC_LightPenHigh=fgetc(SUEF);
        CRTC_LightPenLow=fgetc(SUEF);
        // VIDPROC
        VideoULA_ControlReg=fgetc(SUEF);
        for (int col = 0; col < 16; ++col)
                VideoULA_Palette[col]=fgetc(SUEF)^7; /* Use real ULA values */
        ActualScreenWidth=fget16(SUEF);
        ScreenAdjust=fget32(SUEF);
        CRTCControlReg=fgetc(SUEF);
        TeletextStyle=fgetc(SUEF);
        VideoInit();

//==        if (VideoULA_ControlReg & 2) TeletextEnabled=1; else TeletextEnabled=0;
//==        if (VideoULA_ControlReg & 16) HSyncModifier=8; else HSyncModifier=16;
//==        if (VideoULA_ControlReg & 2) HSyncModifier=12;
        //SetTrigger(99,VideoTriggerCount);
}


/*-------------------------------------------------------------------------*/
//==void SaveVideoState(unsigned char *StateData) {
//==	/* 6845 state */
//==	StateData[0] = CRTCControlReg;
//==	StateData[1] = CRTC_HorizontalTotal;
//==	StateData[2] = CRTC_HorizontalDisplayed;
//==	StateData[3] = CRTC_HorizontalSyncPos;
//==	StateData[4] = CRTC_SyncWidth;
//==	StateData[5] = CRTC_VerticalTotal;
//==	StateData[6] = CRTC_VerticalTotalAdjust;
//==	StateData[7] = CRTC_VerticalDisplayed;
//==	StateData[8] = CRTC_VerticalSyncPos;
//==	StateData[9] = CRTC_InterlaceAndDelay;
//==	StateData[10] = CRTC_ScanLinesPerChar;
//==	StateData[11] = CRTC_CursorStart;
//==	StateData[12] = CRTC_CursorEnd;
//==	StateData[13] = CRTC_ScreenStartHigh;
//==	StateData[14] = CRTC_ScreenStartLow;
//==	StateData[15] = CRTC_CursorPosHigh;
//==	StateData[16] = CRTC_CursorPosLow;
//==	StateData[17] = CRTC_LightPenHigh;
//==	StateData[18] = CRTC_LightPenLow;
//==
//==	/* Video ULA state */
//==	StateData[32] = VideoULA_ControlReg;
//==	for (int col = 0; col < 16; ++col)
//==		StateData[33+col] = VideoULA_Palette[col] ^ 7; /* Use real ULA values */
//==}

/*-------------------------------------------------------------------------*/
//==void RestoreVideoState(unsigned char *StateData) {
//==	/* 6845 state */
//==	CRTCControlReg = StateData[0];
//==	CRTC_HorizontalTotal = StateData[1];
//==	CRTC_HorizontalDisplayed = StateData[2];
//==	CRTC_HorizontalSyncPos = StateData[3];
//==	CRTC_SyncWidth = StateData[4];
//==	CRTC_VerticalTotal = StateData[5];
//==	CRTC_VerticalTotalAdjust = StateData[6];
//==	CRTC_VerticalDisplayed = StateData[7];
//==	CRTC_VerticalSyncPos = StateData[8];
//==	CRTC_InterlaceAndDelay = StateData[9];
//==	CRTC_ScanLinesPerChar = StateData[10];
//==	CRTC_CursorStart = StateData[11];
//==	CRTC_CursorEnd = StateData[12];
//==	CRTC_ScreenStartHigh = StateData[13];
//==	CRTC_ScreenStartLow = StateData[14];
//==	CRTC_CursorPosHigh = StateData[15];
//==	CRTC_CursorPosLow = StateData[16];
//==	CRTC_LightPenHigh = StateData[17];
//==	CRTC_LightPenLow = StateData[18];
//==
//==	/* Video ULA state */
//==	VideoULA_ControlReg = StateData[32];
//==	for (int col = 0; col < 16; ++col)
//==		VideoULA_Palette[col] = StateData[33+col] ^ 7; /* Convert ULA values to colours */
//==
//==	/* Reset the other video state variables */
//==	VideoInit();
//==}

/*-------------------------------------------------------------------------------------------------------------*/
void video_dumpstate(void) {
  int tmp;
  std::cerr << "video:\n";
  std::cerr << "  VideoULA_ControlReg=" << int(VideoULA_ControlReg) << "\n";
  std::cerr << "  VideoULA_Palette=";
  for(tmp=0;tmp<16;tmp++)
    std::cerr << int(VideoULA_Palette[tmp]) << " ";
  std::cerr << "\n  CRTC Control=" << int(CRTCControlReg) << "\n";
  std::cerr << "  CRTC_HorizontalTotal=" << int(CRTC_HorizontalTotal) << "\n";
  std::cerr << "  CRTC_HorizontalDisplayed=" << int(CRTC_HorizontalDisplayed)<< "\n";
  std::cerr << "  CRTC_HorizontalSyncPos=" << int(CRTC_HorizontalSyncPos)<< "\n";
  std::cerr << "  CRTC_SyncWidth=" << int(CRTC_SyncWidth)<< "\n";
  std::cerr << "  CRTC_VerticalTotal=" << int(CRTC_VerticalTotal)<< "\n";
  std::cerr << "  CRTC_VerticalTotalAdjust=" << int(CRTC_VerticalTotalAdjust)<< "\n";
  std::cerr << "  CRTC_VerticalDisplayed=" << int(CRTC_VerticalDisplayed)<< "\n";
  std::cerr << "  CRTC_VerticalSyncPos=" << int(CRTC_VerticalSyncPos)<< "\n";
  std::cerr << "  CRTC_InterlaceAndDelay=" << int(CRTC_InterlaceAndDelay)<< "\n";
  std::cerr << "  CRTC_ScanLinesPerChar=" << int(CRTC_ScanLinesPerChar)<< "\n";
  std::cerr << "  CRTC_CursorStart=" << int(CRTC_CursorStart)<< "\n";
  std::cerr << "  CRTC_CursorEnd=" << int(CRTC_CursorEnd)<< "\n";
  std::cerr << "  CRTC_ScreenStartHigh=" << int(CRTC_ScreenStartHigh)<< "\n";
  std::cerr << "  CRTC_ScreenStartLow=" << int(CRTC_ScreenStartLow)<< "\n";
  std::cerr << "  CRTC_CursorPosHigh=" << int(CRTC_CursorPosHigh)<< "\n";
  std::cerr << "  CRTC_CursorPosLow=" << int(CRTC_CursorPosLow)<< "\n";
  std::cerr << "  CRTC_LightPenHigh=" << int(CRTC_LightPenHigh)<< "\n";
  std::cerr << "  CRTC_LightPenLow=" << int(CRTC_LightPenLow)<< "\n";
}; /* video_dumpstate */

